"""
Copyright 2023 The HuggingFace Team
"""

import json
from dataclasses import dataclass
from typing import Union

import pandas as pd

from autotrain import logger
from autotrain.backend import SpaceRunner
from autotrain.dataset import AutoTrainDataset, AutoTrainDreamboothDataset, AutoTrainImageClassificationDataset
from autotrain.tasks import TASKS
from autotrain.trainers.clm.params import LLMTrainingParams
from autotrain.trainers.dreambooth.params import DreamBoothTrainingParams
from autotrain.trainers.image_classification.params import ImageClassificationParams
from autotrain.trainers.seq2seq.params import Seq2SeqParams
from autotrain.trainers.tabular.params import TabularParams
from autotrain.trainers.text_classification.params import TextClassificationParams


@dataclass
class AutoTrainProject:
    dataset: Union[AutoTrainDataset, AutoTrainDreamboothDataset, AutoTrainImageClassificationDataset]
    job_params: pd.DataFrame

    def __post_init__(self):
        self.token = self.dataset.token
        self.project_name = self.dataset.project_name
        self.username = self.dataset.username
        self.task = self.dataset.task
        if isinstance(self.dataset, AutoTrainDataset):
            self.col_mapping = self.dataset.column_mapping
        self.data_path = f"{self.username}/autotrain-data-{self.project_name}"

        self.backend = self.job_params.loc[0, "backend"]
        if "model_choice" in self.job_params.columns:
            self.model_choice = self.job_params.loc[0, "model_choice"]
        if "param_choice" in self.job_params.columns:
            self.param_choice = self.job_params.loc[0, "param_choice"]

        self.task_id = TASKS.get(self.task)
        self.num_jobs = len(self.job_params)

        if self.task in ("text_multi_class_classification", "text_binary_classification", "seq2seq"):
            self.col_map_text = "autotrain_text"
            self.col_map_target = "autotrain_label"
        if self.task == "lm_training":
            self.col_map_text = "autotrain_text"
        if self.task == "image_multi_class_classification":
            self.col_map_image = "autotrain_image"
            self.col_map_target = "autotrain_label"
        if self.task.startswith("tabular_"):
            self.col_map_id = "autotrain_id"
            _tabular_target_cols = ["autotrain_label"]
            if isinstance(self.col_mapping["label"], str) or len(self.col_mapping["label"]) > 1:
                _tabular_target_cols = [f"autotrain_label_{i}" for i in range(len(self.col_mapping["label"]))]
            self.col_map_target = _tabular_target_cols

        self.spaces_backends = {
            "A10G Large": "spaces-a10gl",
            "A10G Small": "spaces-a10gs",
            "A100 Large": "spaces-a100",
            "T4 Medium": "spaces-t4m",
            "T4 Small": "spaces-t4s",
            "CPU Upgrade": "spaces-cpu",
            "CPU (Free)": "spaces-cpuf",
            "DGX 1xA100": "dgx-a100",
            "DGX 2xA100": "dgx-2a100",
            "DGX 4xA100": "dgx-4a100",
            "DGX 8xA100": "dgx-8a100",
            "Local": "local",
            "EP US-East-1 1xA10g": "ep-aws-useast1-m",
            "EP US-East-1 1xA100": "ep-aws-useast1-xl",
            "EP US-East-1 2xA100": "ep-aws-useast1-2xl",
            "EP US-East-1 4xA100": "ep-aws-useast1-4xl",
            "EP US-East-1 8xA100": "ep-aws-useast1-8xl",
        }

        self.job_params_json = self.job_params.to_json(orient="records")
        logger.info(self.job_params_json)

    def _munge_common_params(self, job_idx):
        _params = json.loads(self.job_params_json)[job_idx]
        _params["token"] = self.token
        _params["project_name"] = f"{self.project_name}-{job_idx}"
        _params["push_to_hub"] = True
        _params["repo_id"] = f"{self.username}/{self.project_name}-{job_idx}"
        _params["data_path"] = self.data_path
        _params["username"] = self.username
        return _params

    def _munge_params_llm(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["model"] = self.model_choice
        _params["text_column"] = self.col_map_text
        _params["prompt_text_column"] = "autotrain_prompt"
        _params["rejected_text_column"] = "autotrain_rejected_text"

        if "trainer" in _params:
            _params["trainer"] = _params["trainer"].lower()

        return _params

    def _munge_params_text_clf(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["model"] = self.model_choice
        _params["text_column"] = self.col_map_text
        _params["target_column"] = self.col_map_target
        _params["valid_split"] = "validation"

        return _params

    def _munge_params_seq2seq(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["model"] = self.model_choice
        _params["text_column"] = self.col_map_text
        _params["target_column"] = self.col_map_target
        _params["valid_split"] = "validation"

        return _params

    def _munge_params_img_clf(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["model"] = self.model_choice
        _params["image_column"] = self.col_map_image
        _params["target_column"] = self.col_map_target
        _params["valid_split"] = "validation"

        return _params

    def _munge_params_tabular(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["id_column"] = self.col_map_id
        _params["target_columns"] = self.col_map_target
        _params["valid_split"] = "validation"

        if len(_params["categorical_imputer"].strip()) == 0 or _params["categorical_imputer"].lower() == "none":
            _params["categorical_imputer"] = None
        if len(_params["numerical_imputer"].strip()) == 0 or _params["numerical_imputer"].lower() == "none":
            _params["numerical_imputer"] = None
        if len(_params["numeric_scaler"].strip()) == 0 or _params["numeric_scaler"].lower() == "none":
            _params["numeric_scaler"] = None

        return _params

    def _munge_params_dreambooth(self, job_idx):
        _params = self._munge_common_params(job_idx)
        _params["model"] = self.model_choice
        _params["image_path"] = self.data_path

        if "weight_decay" in _params:
            _params["adam_weight_decay"] = _params["weight_decay"]
            _params.pop("weight_decay")

        return _params

    def create_spaces(self):
        _created_spaces = []
        for job_idx in range(self.num_jobs):
            if self.task_id == 9:
                _params = self._munge_params_llm(job_idx)
                _params = LLMTrainingParams(**_params)
            elif self.task_id in (1, 2):
                _params = self._munge_params_text_clf(job_idx)
                _params = TextClassificationParams(**_params)
            elif self.task_id in (13, 14, 15, 16, 26):
                _params = self._munge_params_tabular(job_idx)
                _params = TabularParams(**_params)
            elif self.task_id == 25:
                _params = self._munge_params_dreambooth(job_idx)
                _params = DreamBoothTrainingParams(**_params)
            elif self.task_id == 28:
                _params = self._munge_params_seq2seq(job_idx)
                _params = Seq2SeqParams(**_params)
            elif self.task_id == 18:
                _params = self._munge_params_img_clf(job_idx)
                _params = ImageClassificationParams(**_params)
            else:
                raise NotImplementedError
            logger.info(f"Creating Space for job: {job_idx}")
            logger.info(f"Using params: {_params}")
            sr = SpaceRunner(params=_params, backend=self.spaces_backends[self.backend])
            space_id = sr.prepare()
            logger.info(f"Space created with id: {space_id}")
            _created_spaces.append(space_id)
        return _created_spaces

    def create(self):
        if self.backend == "AutoTrain":
            raise NotImplementedError
        if self.backend in self.spaces_backends:
            return self.create_spaces()



import os
import subprocess
import sys
from argparse import ArgumentParser

import torch

from autotrain import logger
from autotrain.backend import SpaceRunner
from autotrain.commands import launch_command

from . import BaseAutoTrainCommand


def run_text_classification_command_factory(args):
    return RunAutoTrainTextClassificationCommand(args)


class RunAutoTrainTextClassificationCommand(BaseAutoTrainCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        arg_list = [
            {
                "arg": "--train",
                "help": "Train the model",
                "required": False,
                "action": "store_true",
            },
            {
                "arg": "--deploy",
                "help": "Deploy the model",
                "required": False,
                "action": "store_true",
            },
            {
                "arg": "--inference",
                "help": "Run inference",
                "required": False,
                "action": "store_true",
            },
            {
                "arg": "--data-path",
                "help": "Train dataset to use",
                "required": False,
                "type": str,
            },
            {
                "arg": "--train-split",
                "help": "Test dataset split to use",
                "required": False,
                "type": str,
                "default": "train",
            },
            {
                "arg": "--valid-split",
                "help": "Validation dataset split to use",
                "required": False,
                "type": str,
                "default": None,
            },
            {
                "arg": "--text-column",
                "help": "Text column to use",
                "required": False,
                "type": str,
                "default": "text",
            },
            {
                "arg": "--target-column",
                "help": "Target column to use",
                "required": False,
                "type": str,
                "default": "target",
            },
            {
                "arg": "--model",
                "help": "Model to use",
                "required": False,
                "type": str,
            },
            {
                "arg": "--lr",
                "help": "Learning rate to use",
                "required": False,
                "type": float,
                "default": 3e-5,
            },
            {
                "arg": "--epochs",
                "help": "Number of training epochs to use",
                "required": False,
                "type": int,
                "default": 1,
            },
            {
                "arg": "--max-seq-length",
                "help": "Maximum number of tokens in a sequence to use",
                "required": False,
                "type": int,
                "default": 128,
            },
            {
                "arg": "--batch-size",
                "help": "Training batch size to use",
                "required": False,
                "type": int,
                "default": 2,
            },
            {
                "arg": "--warmup-ratio",
                "help": "Warmup proportion to use",
                "required": False,
                "type": float,
                "default": 0.1,
            },
            {
                "arg": "--gradient-accumulation",
                "help": "Gradient accumulation steps to use",
                "required": False,
                "type": int,
                "default": 1,
            },
            {
                "arg": "--optimizer",
                "help": "Optimizer to use",
                "required": False,
                "type": str,
                "default": "adamw_torch",
            },
            {
                "arg": "--scheduler",
                "help": "Scheduler to use",
                "required": False,
                "type": str,
                "default": "linear",
            },
            {
                "arg": "--weight-decay",
                "help": "Weight decay to use",
                "required": False,
                "type": float,
                "default": 0.0,
            },
            {
                "arg": "--max-grad-norm",
                "help": "Max gradient norm to use",
                "required": False,
                "type": float,
                "default": 1.0,
            },
            {
                "arg": "--seed",
                "help": "Seed to use",
                "required": False,
                "type": int,
                "default": 42,
            },
            {
                "arg": "--logging-steps",
                "help": "Logging steps to use",
                "required": False,
                "type": int,
                "default": -1,
            },
            {
                "arg": "--project-name",
                "help": "Output directory",
                "required": False,
                "type": str,
            },
            {
                "arg": "--evaluation-strategy",
                "help": "Evaluation strategy to use",
                "required": False,
                "type": str,
                "default": "epoch",
            },
            {
                "arg": "--save-total-limit",
                "help": "Save total limit to use",
                "required": False,
                "type": int,
                "default": 1,
            },
            {
                "arg": "--save-strategy",
                "help": "Save strategy to use",
                "required": False,
                "type": str,
                "default": "epoch",
            },
            {
                "arg": "--auto-find-batch-size",
                "help": "Auto find batch size True/False",
                "required": False,
                "action": "store_true",
            },
            {
                "arg": "--mixed-precision",
                "help": "fp16, bf16, or None",
                "required": False,
                "type": str,
                "default": None,
                "choices": ["fp16", "bf16", None],
            },
            {
                "arg": "--token",
                "help": "Hugging face token",
                "required": False,
                "type": str,
                "default": "",
            },
            {
                "arg": "--push-to-hub",
                "help": "Push to hub True/False. In case you want to push the trained model to huggingface hub",
                "required": False,
                "action": "store_true",
            },
            {
                "arg": "--repo-id",
                "help": "Repo id for hugging face hub",
                "required": False,
                "type": str,
            },
            {
                "arg": "--backend",
                "help": "Backend to use: default or spaces. Spaces backend requires push_to_hub and repo_id",
                "required": False,
                "type": str,
                "default": "default",
            },
            {
                "arg": "--username",
                "help": "Huggingface username to use",
                "required": False,
                "type": str,
            },
            {
                "arg": "--log",
                "help": "Use experiment tracking",
                "required": False,
                "type": str,
                "default": "none",
            },
        ]
        run_text_classification_parser = parser.add_parser(
            "text-classification", description="✨ Run AutoTrain Text Classification"
        )
        for arg in arg_list:
            if "action" in arg:
                run_text_classification_parser.add_argument(
                    arg["arg"],
                    help=arg["help"],
                    required=arg.get("required", False),
                    action=arg.get("action"),
                    default=arg.get("default"),
                )
            else:
                run_text_classification_parser.add_argument(
                    arg["arg"],
                    help=arg["help"],
                    required=arg.get("required", False),
                    type=arg.get("type"),
                    default=arg.get("default"),
                )
        run_text_classification_parser.set_defaults(func=run_text_classification_command_factory)

    def __init__(self, args):
        self.args = args

        store_true_arg_names = [
            "train",
            "deploy",
            "inference",
            "auto_find_batch_size",
            "push_to_hub",
        ]
        for arg_name in store_true_arg_names:
            if getattr(self.args, arg_name) is None:
                setattr(self.args, arg_name, False)

        if self.args.train:
            if self.args.project_name is None:
                raise ValueError("Project name must be specified")
            if self.args.data_path is None:
                raise ValueError("Data path must be specified")
            if self.args.model is None:
                raise ValueError("Model must be specified")
            if self.args.push_to_hub:
                if self.args.repo_id is None:
                    raise ValueError("Repo id must be specified for push to hub")
        else:
            raise ValueError("Must specify --train, --deploy or --inference")

        if not torch.cuda.is_available():
            self.device = "cpu"

        self.num_gpus = torch.cuda.device_count()

        if len(str(self.args.token)) < 6:
            self.args.token = os.environ.get("HF_TOKEN", None)

    def run(self):
        from autotrain.trainers.text_classification.__main__ import train as train_text_classification
        from autotrain.trainers.text_classification.params import TextClassificationParams

        logger.info("Running Text Classification")
        if self.args.train:
            params = TextClassificationParams(
                data_path=self.args.data_path,
                train_split=self.args.train_split,
                valid_split=self.args.valid_split,
                text_column=self.args.text_column,
                target_column=self.args.target_column,
                model=self.args.model,
                lr=self.args.lr,
                epochs=self.args.epochs,
                max_seq_length=self.args.max_seq_length,
                batch_size=self.args.batch_size,
                warmup_ratio=self.args.warmup_ratio,
                gradient_accumulation=self.args.gradient_accumulation,
                optimizer=self.args.optimizer,
                scheduler=self.args.scheduler,
                weight_decay=self.args.weight_decay,
                max_grad_norm=self.args.max_grad_norm,
                seed=self.args.seed,
                logging_steps=self.args.logging_steps,
                project_name=self.args.project_name,
                evaluation_strategy=self.args.evaluation_strategy,
                save_total_limit=self.args.save_total_limit,
                save_strategy=self.args.save_strategy,
                auto_find_batch_size=self.args.auto_find_batch_size,
                mixed_precision=self.args.mixed_precision,
                push_to_hub=self.args.push_to_hub,
                repo_id=self.args.repo_id,
                token=self.args.token,
                username=self.args.username,
                log=self.args.log,
            )

            if self.args.backend.startswith("spaces"):
                logger.info("Creating space...")
                sr = SpaceRunner(
                    params=params,
                    backend=self.args.backend,
                )
                space_id = sr.prepare()
                logger.info(f"Training Space created. Check progress at https://hf.co/spaces/{space_id}")
                sys.exit(0)

            params.save(output_dir=self.args.project_name)
            if self.num_gpus == 1:
                train_text_classification(params)
            else:
                cmd = launch_command(params)

                env = os.environ.copy()
                process = subprocess.Popen(cmd, env=env)
                process.wait()
